/* libthrowable - a C++ library to ease error handling and debugging
 *
 * Copyright (C) Matthias Langer 2006 <mlangc@gmx.at>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef H_LIBTHROWABLE_LIBTHROWABLE
#define H_LIBTHROWABLE_LIBTHROWABLE
#include "libthrowable/system.h"
#include "libthrowable/error.h"
#include "libthrowable/problem.h"
#include "libthrowable/exception.h"
#include "libthrowable/failed_assertion.h"
#include "libthrowable/user_fault_type.h"
#include "libthrowable/trace_point.h"
#include "libthrowable/macros/here.h"
#include "libthrowable/macros/assert.h"
#include "libthrowable/macros/trace.h"

/*! @defgroup exceptions Exception handling
 *  @defgroup debugging Debugging
 *  @defgroup various Various*/

/*!@mainpage
 * 	@section summary Summary
 *		libthrowable is a small library to ease error handling and debugging in C++.
 *	@section rationale Rationale
 *		Error handling and bug fixing are a very important part of programming.
 *		However, what the C++ standard library provides in this area is quite
 *		minimal and does not enforce a particular design. This is where
 *		libthrowable steps in. To help you with error handling, the library
 *		provides a view basic exception types derived from libthrowable::Throwable
 *		that can do much more for you than the standard std::exception classes. 
 *		Deriving new, fully featured exception classes from
 *		the base types that come with the library is very easy thanks to templates.
 *		For debugging, libthrowable::TracePoint together with three macros 
 *		should help you to find out what's going on under the hood.
 *		@section why Why should I use libthrowable?
 *		<ul>
 *			<li>
 *				libthrowable implicitly suggests you a certain exception handling policy that
 *				should lead to clearer code.
 *			</li>
 *
 *			<li>
 *				libthrowable is capable of producing very nice exception messages, tweaked for
 *				the current width of your terminal.
 *			</li>
 *
 *			<li>
 *				libthrowable supports (limited) exception chaining.
 *			</li>
 *
 *			<li>
 *				libthrowable may save you a lot of typing as it provides a few convenience
 *				macros.
 *			</li>
 *
 *			<li>
 *				libthrowable may help you debugging your program/library.
 *			</li>
 *
 *			<li>
 *				libthrowable is designed to be very flexible.
 *			</li>
 *
 *			<li>
 *				libthrowable is well documented.
 *			</li>	
 *
 *			<li>
 *				libthrowable is licensed under the LGPL.
 *			</li>	
 *		</ul>
 *	@section news What's new in version 1.0.0?
 *		A macro called #LIBTHROWABLE_TRACE_VAR has been 
 *		added for your convienience. Besides of this the second
 *		argument of libthrowable::util::optimizeForLineWidth(const std::string&,int)
 *		no longer defaults to libthrowable::util::getTerminalWidth(). You should
 *		use libthrowable::util::optimizeForTerminalWidth(const std::string&)
 *		if you want the old behaviour. Please note, that while <b>libthrowable-1.0.0 is not
 *		API/ABI compatible with libthrowable-0.9.6</b>, in 99.9% of all cases it should
 *		be enough to just recompile depending software.
 *	@section portability Will libthrowable work with my compiler/operating system?
 *		In principle, libthrowable should work in every environment with a reasonable 
 *		modern C++ compiler. However, at the moment it has only been verified that
 *		libthrowable works using the following configurations:
 *      <table>
 *        <tr>
 *          <th>OS</th>
 *          <th>Compilers</th>
 *        </tr>
 *        <tr>
 *          <td>Linux</td>
 *          <td>GNU GCC 3.4, 4.1<br>
 *              Intel ICC 9.1
 *          </td>
 *        </tr>
 *        <tr>
 *          <td>FreeBSD</td>
 *          <td>GNU GCC 3.4</td>
 *        </tr>
 *        <tr>
 *          <td>Cygwin</td>
 *          <td>GNU GCC 3.4, 4.1</td>
 *        </tr>
 *        <tr>
 *          <td>Windows 2000, XP</td>
 *          <td>GNU GCC 3.4 (using DevCpp 4.9.9.2)<br>
 *              Microsoft Visual C++ 8.0<sup>1</sup>
 *          </td>
 *        </tr>
 *      </table>
 *		@note Although libthrowable should work with @em any POSIX compatible system, only the
 *		      tested configurations are listed above. 
 *		@note <sup>1</sup>: libthrowable <em>will</em> compile and work under Microsoft VC++. However, due to a non-conformant
 *		      implementation of <code>std::exception</code> (std::exception has a non-explicit
 *		      conversion constructor from <code>const char*</code> in VC++), there are 
 *		      ambiguities in several constructors of libthrowable classes, as the compiler is unable
 *		      to decide whether to use <code>std::string</code> or <code>std::exception</code>. There's a bug report for this
 *		      issue (see http://connect.microsoft.com/VisualStudio/feedback/ViewFeedback.aspx?FeedbackID=200471) which says that
 *		      this issue will be resolved "in an upcoming release of Visual Studio" (whatever that means)...
 *		      Practically this means that the following code is invalid in VC++:
 *		@code
 * // should work with any reasonable C++ implementation; doesn't work with VC++ ...
 * throw libthrowable::Error(LIBTHROWABLE_HERE, "what the hell ...");
 * 		@endcode
 *      The problem is that the compiler is unable to decide whether to construct an
 *      @c std::string or @c std::exception from the <code>const char*</code>.
 *
 *		@warning
 *		When building the library under Windows, make sure that
 *		LIBTHROWABLE_BUILD and (unless you're building static libs) LIBTHROWABLE_DLL
 *		are defined. A project file (libthrowable.dev) is included in the tarball
 *		which should require no further configuration. Also note that, if you're
 *		linking against a DLL file, it is required that you define LIBTHROWABLE_DLL
 *		to <code>__dllimport</code> the neccessary symbols.
 *	@section download Where do I get libthrowable?
 *		You can download libthrowable from http://sourceforge.net/projects/libthrowable or
 *		obtain it directly from the subversion repository at 
 *		@htmlonly 
 *			https://svn.sourceforge.net/svnroot/libthrowable. 
 *		@endhtmlonly
 *	@section tutorial Tutorial
 *		libthrowable is a small and yet simple library and consists mainly of
 *		exception classes. The most important ones are:
 *		<ul>
 *			<li> 
 *				libthrowable::Throwable: this is the base class for all exception classes. It cannot 
 *				be instantiated directly.
 *			</li>
 *			
 *			<li> 
 *				libthrowable::Error: represents a serious program error and should normally terminate
 *				the application.
 *			</li>
 *			
 *			<li>
 *				libthrowable::Problem: as the name says, this class represents a problem. It may or may not
 *				be recoverable.
 *			</li>
 *
 *			<li>
 *				libthrowable::Exception: this class represents a recoverable problem. Thus, Exception objects should
 *				always be caught and handled.
 *			</li>
 *
 *			<li>
 *				libthrowable::UserFault: this class represents an exception object that it is aimed at the
 *				end user. It is derived from libthrowable::Exception.
 *			</li>
 *
 *			<li>
 *				libthrowable::TracePoint: for your debugging purposes.
 *			</li>	
 *		</ul>
 *		<b>Note:</b> All examples mentioned here can be found in the examples directory. You do not even need to
 *		compile them yourself if you pass "--enable-examples" to configure.
 *
 * 		@subsection exception Exception handling
 *			Here is a simple example that shows the most basic usage of libthrowable. If you have pkg-config, g++ and
 *			the library is properly installed, you should be able to build the source below with
 *			"g++ filename.cc `pkg-config libthrowable-1.0 --cflags --libs` -o executable".
 *			@include basic_usage.cc
 *			As you see, usage of the library is straight forward.
 *			@subsection userFault A closer look at libthrowable::UserFault
 *				As stated before, libthrowable::UserFault is an exception type that carries a message aimed directly at the
 *				user of the program you are working on. Figuratively speaking, a libthrowable::UserFault should be caught and
 *				handled by the user of your software, while coping with libthrowable::Problem and libthrowable::Exception is
 *				your responsibility as programmer (we didn't mention libthrowable::Error
 *				here because objects of this type should terminate your program). To emphasize this difference, libthrowable::UserFault provies the
 *				additional method libthrowable::UserFault::getMessage() that returns exactly the message you passed to the
 *				constructor of the object without additional information like libthrowable::Throwable::toString() does. Besides,
 *				operator<<(std::ostream&, const libthrowable::UserFault&) is implemented using
 *				libthrowable::UserFault::getMessage() while operator<<(std::ostream&, const libthrowable::Throwable&) calls
 *				libthrowable::Throwable::toString().
 *			@subsubsection deriving Deriving your own exception types
 *				You may have noticed, that libthrowable itself only provides very few exception types. That is
 *				in fact not because the authors of the library have been lazy, but to keep things flexible. Instead,
 *				you should derive your own types as needed using one of the template classes
 *				libthrowable::ThrowableType or libthrowable::UserFaultType. This is illustrated below:
 *				@include deriving_your_own_types.cc
 *			@subsubsection exNotes Notes
 *				As already stated, libthrowable is capable to tweak it's output to the current width of your terminal. For this purpose,
 *				libthrowable::Throwable::toString(int) is provided. Besides, operator<<(std::ostream&, const libthrowable::Throwable&)
 *				calls libthrowable::Throwable::toString(int) with default arguments if you work with cerr or cout. Thus, to cut a long story
 *				short, libthrowable will tweak it's output for your terminal by default (unless you terminal is too small).		
 *		@subsection debugging Using libthrowable for debugging
 *			At first, libthrowable provides #LIBTHROWABLE_ASSERT that just behaves like the normal assert macro, but throws
 *			a libthrowable::FailedAssertion object if the assertion did not hold. You can deactivate #LIBTHROWABLE_ASSERT by
 *			defining LIBTHROWABLE_NDEBUG. @n
 *			However, the main debugging feature of this library is not #LIBTHROWABLE_ASSERT but libthrowable::TracePoint.
 *			For those who want to save some typing, there are the convenience macros
 *			#LIBTHROWABLE_TRACE, #LIBTHROWABLE_TRACE_WITH_MSG and #LIBTHROWABLE_TRACE_VAR. Calling
 *			@code LIBTHROWABLE_TRACE(tpX); @endcode
 *			is equivalent to
 *			@code libthrowable::TracePoint tpX(LIBTHROWABLE_HERE); @endcode
 *			,
 *			@code LIBTHROWABLE_TRACE_WITH_MSG(tpX, msg); @endcode
 *			is the same as
 *			@code libthrowable::TracePoint tpX(LIBTHROWABLE_HERE, msg); @endcode
 *			and
 *			@code LIBTHROWABLE_TRACE_VAR(tpX, var); @endcode
 *			is the same as
 *			@code libthrowable::TracePoint tpX(LIBTHROWABLE_HERE, var, "var name"); @endcode
 *			These macros can be deactivated for production code by defining LIBTHROWABLE_NDEBUG or LIBTHROWABLE_NTRACE. After this
 *			short introduction, let's see what libthrowable::TracePoint can do for you. Compile and run the following code:
 *			@include trace_point_usage.cc
 *			After running this program you should have a clue what libthrowable::TracePoint is for and how it can help
 *			you with debugging. Note, that again, the output is tweaked for your terminal width unless your terminal is unreasonable
 *			small.@n@n
 *			libthrowable::TracePoint also behaves well in a multi-threaded environment, using either POSIX threads (pthreads) or the
 *			native Windows Thread API. Note that when using pthreads you have to enable pthreads support before compilation.
 *			This can be done by passing "--enable-pthreads" to configure, or, if you are using the gentoo ebuild, by setting the "threads"
 *			useflag. Compile and run the following example to see how the library copes with pthreads:
 *			@include trace_point_with_pthreads.cc
 *		@subsection together Putting it all together
 *			This example shows you all the above together:
 *			@include putting_all_together.cc
 *		@subsection next Where to go from here
 *			You should now know enough to use the library. Consult the reference manual for more detailed descriptions.
 *	@section misc Miscellaneous
 *		@subsection api Is libthrowable API/ABI stable?
 *			libthrowable-1.0.0 is the first stable release of the 1.0.x series that will stay API/ABI compatible with 1.0.0.
 *		@subsection nextRel What is planned for the next relase?
 *			At the moment only bugfixes are planned.
 *		@subsection bugs Where to report bugs?
 *			You should report bugs at http://sourceforge.net/projects/libthrowable or write an email directly
 *			to \<mlangc AT gmx DOT at\>.
 */
#endif
