#include "internal.h"
#include "libthrowable/system.h"
#include "libthrowable/macros/here.h"
#include "libthrowable/util/util.h"
#include <string>
#include <sstream>
#include <vector>
#include <iostream>
#include <limits>
#if defined LIBTHROWABLE_WIN32
#include <windows.h>
// might poison the code below
#undef max
#endif

using namespace std;
namespace libthrowable {
namespace internal {
/////////////
//local stuff
/////////////
namespace {
namespace local {
	string tabsToSpaces(const string& str, unsigned spaces) throw();
	long sizeTToLong(string::size_type st) throw();

	string tabsToSpaces(const string& str, unsigned spaces) throw()
	{
		string ret(str);
		string::size_type pos=0;
		while(true)
		{
			pos = ret.find('\t', pos);
			if(pos == string::npos)
				break;
			ret.replace(pos, 1, spaces, ' ');
		}
		return ret;
	}

	long sizeTToLong(string::size_type st) throw()
	{
		if(st <= static_cast<string::size_type>(numeric_limits<long>::max()))
			return static_cast<long>(st);
		stringstream err;
		err << "    could not safely cast " << st << " to long due to limited value range." << endl;
		internal::emergencyAbort(err.str(), LIBTHROWABLE_HERE);
	}
	#if defined LIBTHROWABLE_WIN32
	class PVoid
	{
		public:
		PVoid() throw()
		: _ptr(0) {}

		~PVoid() throw()
		{
			if(_ptr)
				LocalFree(_ptr);
		}

		LPTSTR toLptstr() throw()
		{
			return reinterpret_cast<LPTSTR>(&_ptr);
		}

		char* toCstring() throw()
		{
			return static_cast<char*>(_ptr);
		}

		private:
		PVOID _ptr;
	};

	string winStrerror(DWORD errnum) throw()
	{
		PVoid buffer;
		DWORD flags = FORMAT_MESSAGE_ALLOCATE_BUFFER |
			      FORMAT_MESSAGE_FROM_SYSTEM |
			      FORMAT_MESSAGE_IGNORE_INSERTS;
		FormatMessage(flags, NULL, errnum, 0, buffer.toLptstr(), 0, NULL);
		if(buffer.toLptstr() && buffer.toCstring())
			return string(buffer.toCstring());
		return "";
	}
	#endif


}}
//////////////////////////
//StringAlignmentOptimizer
/////////////////////////
StringAlignmentOptimizer::StringAlignmentOptimizer(int width) throw()
: _width(width), _maxIndent(width/2+1) {}

string StringAlignmentOptimizer::optimize(const string& str) throw()
{
	if(_width < 40 || str.empty())
		return str;
	string ret(local::tabsToSpaces(str, 8));
	ret = adjustIndent(ret);
	ret = adjustLineBreaks(ret);
	return ret;
}

long StringAlignmentOptimizer::getIndent(const string& line) throw()
{
	if(line.empty())
		return -1;
	
	string::size_type ret = line.find_first_not_of(' ');
	if(ret != string::npos)
		return local::sizeTToLong(ret);
	return -1;
}

string::size_type StringAlignmentOptimizer::getLinebreakPosNear(const string& line, string::size_type pos) throw()
{
	if(pos >= line.size())
		return string::npos;
	const long iIndent = getIndent(line);
	if(iIndent < 0)
		return string::npos;
	const string::size_type indent = static_cast<string::size_type>(iIndent);
	//if possible, break after one of the following characters:
	//" .,;"
	string::size_type ret = line.find_last_of(" ,;.", pos-1);
	if(ret != string::npos && ret > indent)
		return ret;
	//maybe we can find a break point by looking after these chars:
	//"?<>{}()=/\|&+-#*:"
	ret = line.find_last_of("?<>(){}=/\\|&+-#:*", pos-1);
	if(ret != string::npos && ret > indent)
		return ret;
	//just break line at pos ...
	return pos;
}

long StringAlignmentOptimizer::getVirtualIndent(const string& line) throw()
{
	const long lineIndent = getIndent(line);

	if(lineIndent == -1)
		return getCurrentVirtualIndent();

	long realIndent = lineIndent - getCurrentVirtualIndent();
	if(realIndent < 0)
	{
		while(true)
		{
			_vIndentStack.pop();
			realIndent = lineIndent - getCurrentVirtualIndent();
			if(realIndent >= 0)
				break;
		}
	}
	if(realIndent < _maxIndent)
		return getCurrentVirtualIndent();
	_vIndentStack.push(getCurrentVirtualIndent() + realIndent);
	return getCurrentVirtualIndent();
}

string StringAlignmentOptimizer::adjustIndent(const string& str) throw()
{
	string ret;
	istringstream istr(str);
	long lastVirtualIndent = getCurrentVirtualIndent();
	bool firstLine = true;
	while(istr.good())
	{
		if(firstLine)
			firstLine = false;
		else
			ret += "\n";

		string line;
		getline(istr, line);
		if(line.empty())
		{
			ret += "\n";
			continue;
		}

		long virtualIndent = getVirtualIndent(line);
		long virtualIndentDiff = virtualIndent - lastVirtualIndent;
		if(virtualIndentDiff > 0)
			ret += "--" + util::stringify(virtualIndentDiff) + "--spaces-->\n";
		else if(virtualIndentDiff < 0)
			ret += "<--" + util::stringify(-virtualIndentDiff) + "-spaces--\n";
		ret += line.substr(static_cast<string::size_type>(virtualIndent));
		lastVirtualIndent = virtualIndent;
	}
	return ret;
}

string StringAlignmentOptimizer::adjustLineBreaks(const string& str) const throw()
{
	string ret;
	istringstream istr(str);
	bool firstLine = true;
	while(istr.good())
	{
		if(firstLine)
			firstLine = false;
		else
			ret += "\n";

		vector<string> lines(1, string());
		getline(istr, lines[0]);
		if(!lines.back().empty())
		{
			long indent = getIndent(lines.back());
			if(indent != -1)
			{
				string prefix(indent, ' ');
				while(true)
				{
					string::size_type breakAt = getLinebreakPosNear(lines.back(), _width);
					if(breakAt == string::npos)
						break;

					if(local::sizeTToLong(indent) > _width)
					{
						stringstream err;
						err << "    breakAt: " << breakAt << endl;
						err << "    indent: " << indent << endl;
						err << "    lines.size(): " << lines.size() << endl;
						err << endl;
						err << str << endl;
						emergencyAbort(err.str(), LIBTHROWABLE_HERE);
					}

					string tmp(lines.back());
					lines.back().erase(breakAt);
					tmp = tmp.substr(breakAt);
					
					while(!tmp.empty() && tmp[0] == ' ')
						tmp.erase(0, 1);
					lines.push_back(prefix + tmp);
				}
			}
		}
		for(string::size_type i=0; i != lines.size(); ++i)
		{
			if(i != 0)
				ret += "\n";
			ret += lines[i];
		}
	}
	return ret;
}

//////////
//funcions
//////////
void emergencyAbort(const string& errmsg, const char* filename, int line, const char* function, int errnum) throw()
{
	const string title("libthrowable - internal library error");
	#if defined LIBTHROWABLE_WIN32
	stringstream err;
	#else
	ostream& err = cerr;
	#endif
	err << title << endl;
	err << "  where:" << endl;
	err << "    " << filename << ":" << line << " in " << function << endl;
	err << "  why:" << endl;
	err << errmsg << endl;
	if(errnum)
	{
		err << "  system error:" << endl;
		#if defined LIBTHROWABLE_WIN32
		err << "    " << local::winStrerror(errnum) << endl;
		#else
		err << "    " << strerror(errnum) << endl;
		#endif
	}
	err << "please submit a full bug report" << endl;
	err << "aborting..." << endl;
	#if defined LIBTHROWABLE_WIN32
	// display an error dialog AND stderr output when using windows
	cerr << err.str() << endl;
	MessageBox(NULL, err.str().c_str(), title.c_str(), MB_OK | MB_ICONERROR | MB_APPLMODAL);
	#endif
	abort();
}
}}//namespaces
