/* libthrowable - a C++ library to ease error handling and debugging
 *
 * Copyright (C) Matthias Langer 2006 <mlangc@gmx.at>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef H_LIBTHROWABLE_TRACE
#define H_LIBTHROWABLE_TRACE
#include <libthrowable/macros/here.h>
#include <libthrowable/trace_point.h>

#ifndef LIBTHROWABLE_DISABLE_MACROS

#ifdef LIBTHROWABLE_NDEBUG
#define LIBTHROWABLE_NTRACE
#endif //LIBTHROWABLE_NDEBUG

#ifndef LIBTHROWABLE_NTRACE
/** Initializes a libthrowable::TracePoint named 'tracePointName'.
 * @ingroup debugging
 * @param tracePointName
 * 	The object's name.
 * @warning
 * 	tracePointName must not collide with other visible names.
 * @note
 * 	define LIBTHROWABLE_NTRACE or LIBTHROWABLE_NDEBUG to remove the effects of
 * 	this macro for production code.*/
#define LIBTHROWABLE_TRACE(tracePointName) \
	libthrowable::TracePoint tracePointName(LIBTHROWABLE_HERE)

/** Initializes a libthrowable::TracePoint named 'tracePointName' with an additional message.
 * @ingroup debugging
 * @param tracePointName
 * 	The object's name.
 * @param message
 * 	The message you wish to pass. Should be convertable to std::string.
 * @warning
 * 	tracePointName must not collide with other visible names.
 * @note
 * 	define LIBTHROWABLE_NTRACE or LIBTHROWABLE_NDEBUG to remove the effects of
 * 	this macro for production code.*/
#define LIBTHROWABLE_TRACE_WITH_MSG(tracePointName, message) \
	libthrowable::TracePoint tracePointName(LIBTHROWABLE_HERE, (message))

/** Initializes a libthrowable::TracePoint named 'tracePointName' with a message showing
 *  a variable and its value.
 *  @ingroup debugging
 *
 *  This small example shows how to use this macro:
 * @code
 *  int i = 1;
 *  LIBTHROWABLE_TRACE_VAR(tp1, i); // same as
 *                                  // LIBTHROWABLE_TRACE_WITH_MSG(tp1, "i = " + libthrowable::util::stringify(i));
 * @endcode
 * @param tracePointName
 *  	The object's name.
 * @param variable
 *  	A variable you want to trace. Note that it must fullfill all requirements for
 *  	libthrowable::util::stringify().
 * @warning
 * 	tracePointName must not collide with other visible names.
 * @note
 * 	define LIBTHROWABLE_NTRACE or LIBTHROWABLE_NDEBUG to remove the effects of
 * 	this macro for production code.
 * @since libthrowable-1.0.0	
 */
#define LIBTHROWABLE_TRACE_VAR(tracePointName, variable) \
	libthrowable::TracePoint tracePointName(LIBTHROWABLE_HERE, (variable), #variable)

#else //#ifndef LIBTHROWABLE_NTRACE
#define LIBTHROWABLE_TRACE(tracePointName) \
	do { } while(0)
#define LIBTHROWABLE_TRACE_WITH_MSG(tracePointName, message) \
	do { } while(0)

#endif //LIBTHROWABLE_NTRACE
#endif //LIBTHROWABLE_DISABLE_MACROS
#endif //H_LIBTHROWABLE_TRACE
