/* libthrowable - a C++ library to ease error handling and debugging
 *
 * Copyright (C) Matthias Langer 2006 <mlangc@gmx.at>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef H_LIBTHROWABLE_OS
#define H_LIBTHROWABLE_OS

#if defined WIN32 || defined _WIN32 || defined _WIN32_WCE || defined __WIN32__
#  define LIBTHROWABLE_WIN32
#else //assume POSIX compatible
/**
 * Defined if using a POSIX compatible operating system.
 */
#  define LIBTHROWABLE_POSIX
#endif

// for Cygwin, both LIBTHROWABLE_POSIX and LIBTHROWABLE_CYGWIN
// should be defined.
#if defined __CYGWIN__ || defined __CYGWIN32__
#  define LIBTHROWABLE_CYGWIN
#endif

#ifdef LIBTHROWABLE_DLL
#  ifdef LIBTHROWABLE_BUILD
#    ifndef _WINDLL
#      define LIBTHROWABLE_API __declspec(dllexport)
#    else
#      define LIBTHROWABLE_API
#    endif
#  else
#    define LIBTHROWABLE_API __declspec(dllimport)
#  endif //LIBTHROWABLE_BUILD
#else
#  define LIBTHROWABLE_API
#endif //LIBTHROWABLE_DLL

// ensure that the doxygen preprocessor
// defines some macros
#ifdef DOXYGEN_SHOULD_SKIP_THIS
#  define LIBTHROWABLE_DLL
/** 
 * Defined if using Windows.
 */
#  define LIBTHROWABLE_WIN32
/**
 * Defined if using the Cygwin POSIX emulation.
 */
#  define LIBTHROWABLE_CYGWIN

// undefine them to ensure that the doxygen
// preprocessor continues normally

#undef LIBTHROWABLE_DLL
#undef LIBTHROWABLE_WIN32
#undef LIBTHROWABLE_CYGWIN
#endif

/**
 * Marks a function as non-returning.
 *
 * @warning
 * Usage is slightly different compared to
 * GCC's __attribute__ extension.
 * @code
 * // this will work for GCC, but not MSVC++
 * void foo() LIBTHROWABLE_NORETURN;
 * // this is uglier, but works for both (darn!)
 * void LIBTHROWABLE_NORETURN foo();
 * @endcode
 */
#if defined __GNUC__
#define LIBTHROWABLE_NORETURN __attribute__((noreturn))
#elif defined _MSC_VER
#define LIBTHROWABLE_NORETURN __declspec(noreturn)
#else
#define LIBTHROWABLE_NORETURN
#endif


/**
 * @def LIBTHROWABLE_API
 * @brief Windows specific macro for building DLLs.
 * This macro only applies to Windows when building DLLs.
 * It is defined to __declspec(dllexport) while building
 * libthrowable (LIBTHROWABLE_BUILD must be defined when
 * building the library from source).
 * If including a header, LIBTHROWABLE_API will be expanded to __declspec(dllimport).
 */

/** 
 * @def LIBTHROWABLE_DLL
 * 
 * If using Windows, and you're either compiling or
 * using libthrowable as a DLL, define this to ensure
 * proper DLL imports and exports.
 * 
 * @warning In some cases linking against a DLL without 
 *          defining LIBTHROWABLE_DLL will work. However,
 *          ensure that you always define this when using
 *          libthrowable as a dll as unexpected linkage
 *          errors might occur.
 */
#endif //H_LIBTHROWABLE_OS
