/* libthrowable - a C++ library to ease error handling and debugging
 *
 * Copyright (C) Matthias Langer 2006 <mlangc@gmx.at>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "throwable.h"
#include "util/util.h"
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <iostream>
using namespace libthrowable::util;
using namespace std;

namespace libthrowable {
const char* Throwable::what() const throw() 
{ return _string.c_str(); }

string Throwable::what(unsigned indent) const throw()
{ return addIndent(_string, indent); }

string Throwable::toString(int lineWidth) const throw()
{ return optimizeForLineWidth(_string, lineWidth); }

Throwable::~Throwable() throw() {}

Throwable::Throwable(const string& msg, const DontManipulateStr&) throw()
: _string(msg) {}

string& Throwable::getInternalString() throw()
{ return _string; }	

string 
Throwable::internalStrFromFilenameAndLinenumber(
	const string& className, const char *filename, int linenumber) throw()
{ return className + ":\n" + addIndent(strFromFilenameAndLinenumber(filename, linenumber), 2); }

string
Throwable::internalStrFromFilenameLinenumberAndFunction(
	const string& className, const char *filename, int linenumber, const char *function) throw()
{ return className + ":\n" + addIndent(strFromFilenameLinenumberAndFunction(filename, linenumber, function), 2); }

string
Throwable::internalStrFromFilenameLinenumberAndMessage(
	const string& className, const char *filename, int linenumber, const string& msg) throw()
{ return className + ":\n" + addIndent(strFromFilenameLinenumberAndMessage(filename, linenumber, msg), 2); }

string
Throwable::internalStrFromFilenameLinenumberFunctionAndMessage(
	const string& className, const char *filename, int linenumber, const char *function, const string& msg) throw()
{ return className + ":\n" + addIndent(strFromFilenameLinenumberFunctionAndMessage(filename, linenumber, function, msg), 2); }	

string
Throwable::internalStrFromCause(
	const string& className, const exception& cause) throw()
{ return className + ":\n" + addIndent(strFromCause(cause), 2); }

string
Throwable::internalStrFromCauseAndMessage(
	const string& className, const exception& cause, const string& message) throw()
{ return className + ":\n" + addIndent(strFromCauseAndMessage(cause, message), 2); }

string
Throwable::internalStrFromFilenameLinenumberAndCause(
	const string& className, const char *filename, int linenumber, const exception& cause) throw()
{ return className + ":\n" + addIndent(strFromFilenameLinenumberAndCause(filename, linenumber, cause), 2); }

string
Throwable::internalStrFromFilenameLinenumberFunctionAndCause(
	const string& className, const char *filename, int linenumber, const char *function, const exception& cause) throw()
{ return className + ":\n" + addIndent(strFromFilenameLinenumberFunctionAndCause(filename, linenumber, function, cause), 2); }

string
Throwable::internalStrFromFilenameLinenumberFunctionCauseAndMessage(
	const string& className, const char *filename, int linenumber, 
	const char *function, const exception& cause, const string& msg) throw()
{ 
	return className + ":\n" + 
		addIndent(
			strFromFilenameLinenumberFunctionCauseAndMessage(
				filename, linenumber, function, cause, msg),
			2);
}

string Throwable::strFromFilenameAndLinenumber(const char *filename, int linenumber) throw()
{ return "where: " + string(filename) + ":" + stringify(linenumber); }

string Throwable::strFromFilenameLinenumberAndFunction(const char *filename, int linenumber, const char *function) throw() 
{ return strFromFilenameAndLinenumber(filename, linenumber) + " in\n  " + function; }

string 
Throwable::strFromFilenameLinenumberAndMessage(
	const char *filename, int linenumber, const string& msg) throw()
{
	string ret(strFromFilenameAndLinenumber(filename, linenumber) + "\nmessage:\n");
	ret += addIndent(msg, 2);
	return ret;
}

string
Throwable::strFromFilenameLinenumberFunctionAndMessage(
	const char *filename, int linenumber, const char *function, const string& msg) throw()
{
	string ret(strFromFilenameLinenumberAndFunction(filename, linenumber, function) + "\nmessage:\n");
	ret += addIndent(msg, 2);
	return ret;
}

string Throwable::strFromCause(const exception& cause) throw()
{ return string("cause:\n") + addIndent(cause.what(), 2); }

string Throwable::strFromCauseAndMessage(
	const exception& cause, const string& msg) throw()
{ return "message:\n" + addIndent(msg, 2) + "\n" + strFromCause(cause); }	

string 
Throwable::strFromFilenameLinenumberAndCause(
	const char *filename, int linenumber, const exception& cause) throw()
{ return strFromFilenameAndLinenumber(filename, linenumber) + "\n" + strFromCause(cause); }

string
Throwable::strFromFilenameLinenumberFunctionAndCause(
	const char *filename, int linenumber, const char *function, const exception& cause) throw()
{ return strFromFilenameLinenumberAndFunction(filename, linenumber, function) + "\n" + strFromCause(cause); }

string
Throwable::strFromFilenameLinenumberFunctionCauseAndMessage(
		const char *filename, int linenumber, const char *function,
		const exception& cause, const string& msg) throw()
{ 
	string ret(strFromFilenameLinenumberAndFunction(filename, linenumber, function)+ "\n");
	ret += "message:\n" + addIndent(msg, 2) + "\n";
	ret += strFromCause(cause);
	return ret;
}

ostream& operator<<(ostream& os, const Throwable& throwable)
{
	if(&os == &cerr || &os == &cout)
		os << throwable.toString(util::getTerminalWidth());
	else
		os << throwable.toString(120);
	return os;
}
}//namespace

	
