/* libthrowable - a C++ library to ease error handling and debugging
 *
 * Copyright (C) Matthias Langer 2006 <mlangc@gmx.at>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef H_LIBTHROWABLE_THROWABLE
#define H_LIBTHROWABLE_THROWABLE
#include <stdexcept>
#include <ostream>
#include "util/util.h"
#include "system.h"

/** All classes of libthrowable can be found here.*/
namespace libthrowable {

/*! @addtogroup exceptions
 *  @{*/

/** Base class for all exception types in libthrowable.
 * @note
 * 	This class can not be instantiated directly.*/
class LIBTHROWABLE_API Throwable : public std::exception
{
	public:
	/** Inherited from std::exception.
	 * @see toString()
	 * @see what()*/
	virtual const char* what() const throw();

	/** The same like what(), but with additional indent.*/
	virtual std::string what(unsigned indent) const throw();

	/** Converts a Throwable object to a string tweaked for a given line width.
	 * @param lineWidth
	 * 	the maximal line width the output of this method should be tweaked for. 
	 * 	For linewidth < 40, no tweaking is done.
	 * @since libthrowable-0.9.1*/	
	virtual std::string toString(int lineWidth = libthrowable::util::getTerminalWidth()) const throw();

	virtual ~Throwable() throw();

	protected:
	/** Dummy class.
	 * @see Throwable(const std::string&, const DontManipulateStr&)*/
	class DontManipulateStr {};

	/** Use this constructor when deriving from this class.*/
	Throwable(const std::string& msg, const DontManipulateStr& dummy) throw();

	/** Use this one if you want to manipulate the stored string from a derived classes.
	 * @return
	 * 	Basically a reference to the return value of what().*/
	std::string& getInternalString() throw(); 

	/** Convenience function to be used in constructors of derived classes.*/
	static std::string
	internalStrFromFilenameAndLinenumber(
		const std::string& className, const char *filename, int linenumber) throw();

	/** Convenience function to be used in constructors of derived classes.*/
	static std::string
	internalStrFromFilenameLinenumberAndFunction(
		const std::string& className, const char *filename, int linenumber, const char *function) throw();
	
	/** Convenience function to be used in constructors of derived classes.*/
	static std::string
	internalStrFromFilenameLinenumberAndMessage(
		const std::string& className, const char *filename, int linenumber, const std::string& message) throw();

	/** Convenience function to be used in constructors of derived classes.*/
	static std::string
	internalStrFromFilenameLinenumberFunctionAndMessage(
		const std::string& className, const char *filename, int linenumber, const char *function, const std::string& message) throw();

	/** Convenience function to be used in constructors of derived classes.*/
	static std::string
	internalStrFromCause(
		const std::string& className, const std::exception& cause) throw();

	/** Convenience function to be used in constructors of derived classes.
	 * @since libthrowable-0.9.2*/
	static std::string
	internalStrFromCauseAndMessage(
		const std::string& className, const std::exception& cause, const std::string& message) throw();	

	/** Convenience function to be used in constructors of derived classes.*/
	static std::string
	internalStrFromFilenameLinenumberAndCause(
		const std::string& className, const char *filename, int linenumber, const std::exception& cause) throw();

	/** Convenience function to be used in constructors of derived classes.*/
	static std::string
	internalStrFromFilenameLinenumberFunctionAndCause(
		const std::string& className, const char *filename, int linenumber, const char *function, const std::exception& cause) throw();

	/** Convenience function to be used in constructors of derived classes.
	 * @since libthrowable-0.9.2*/
	static std::string
	internalStrFromFilenameLinenumberFunctionCauseAndMessage(
		const std::string& className, const char *filename, int linenumber, const char *function,
		const std::exception& cause, const std::string& msg) throw();

	/** Returns 'where: \<filename\>:\<linenumber\>'.*/
	static std::string 
	strFromFilenameAndLinenumber(
		const char *filename, int linenumber) throw();

	/** Returns 'where:\\n  \<filename\>:\<linenumber\> in\\n    \<function\>'.*/
	static std::string 
	strFromFilenameLinenumberAndFunction(
		const char *filename, int linenumber, const char *function) throw();

	/** Returns 'where: \<filename\>:\<linenumber\> \\nmessage:\\n  \<msg\>'.*/
	static std::string 
	strFromFilenameLinenumberAndMessage(
		const char *filename, int linenumber, const std::string& msg) throw();

	/** Returns 'strFromFilenameAndLinenumber() + \\nmessage:\\n  \<msg\>'.*/
	static std::string
	strFromFilenameLinenumberFunctionAndMessage(
		const char *filename, int linenumber, const char *function, const std::string& msg) throw();

	/** Returns 'cause:\\n \<cause.what()\>'.*/
	static std::string strFromCause(const std::exception& cause) throw();

	/** Returns 'cause:\\n  \<cause.what()\>\\nmessage\\n  \<msg\>'.  
	 * @since libthrowable-0.9.2*/
	static std::string strFromCauseAndMessage(const std::exception& cause, const std::string& msg) throw();

	/** Returns 'where: \<filename\>:\<linenumber\>\\ncause:\\n  \<cause.what()\>'*/
	static std::string
	strFromFilenameLinenumberAndCause(
		const char *filename, int linenumber, const std::exception& cause) throw();

	/** Returns 'where: \<filename\>:\<linenumber\> in\\n    \<function\>\\ncause:\\n  \<cause.what()\>'*/
	static std::string
	strFromFilenameLinenumberFunctionAndCause(
		const char *filename, int linenumber, const char *function, const std::exception& cause) throw();

	/** Returns 'where: \<filename\>:\<linenumber\> in\\n  \<function\>\\nmessage:\\n \<msg\>\\ncause:\\n  \<cause.what()\>'
	 * @since libthrowable-0.9.2*/
	static std::string
	strFromFilenameLinenumberFunctionCauseAndMessage(
		const char *filename, int linenumber, const char *function, 
		const std::exception& cause, const std::string& msg) throw();	

	private:
	std::string _string;
};

/** Writes throwable.toString() to os.
 * @see Throwable::toString() const*/
std::ostream& LIBTHROWABLE_API operator<<(std::ostream& os, const Throwable& throwable);

/*! @}*/
}//namespace
#endif
