/* libthrowable - a C++ library to ease error handling and debugging
 *
 * Copyright (C) Matthias Langer 2006 <mlangc@gmx.at>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef H_LIBTHROWABLE_THROWABLE_TYPE
#define H_LIBTHROWABLE_THROWABLE_TYPE
#include "throwable.h"
#include "util/util.h"
namespace libthrowable {

/** Use this template to derive new throwable classes within seconds.
 * @ingroup exceptions
 * @param ThrowableT
 * 	a type that has a constructor that is compatible with
 * 	ThrowableT::ThrowableT(const std::string&, const ThrowableT::DontManipulateStr&). 
 * 	Most likely this will be a class like Throwable, Exception or Error.
 * @param TagT
 *  a tag type with a static name() method, that returns a (const) string
 *  (reference).
 * @note
 * 	Suppose you want to have a new Error class named Meltdown. You should
 * 	do that like this:
 * 	@code
 * 		#include <libthrowable/error.h>
 * 		using namespace libthrowable;
 * 		struct MeltdownTag
 * 		{
 * 			static const std::string& name()
 * 			{
 * 				static const std::string ret("Meltdown");
 * 				return ret;
 * 			}
 * 		};
 *		typedef ThrowableType<Error, MeltdownTag> Meltdown;
 *  @endcode
 * @see UserFaultType*/ 
template<class ThrowableT, class TagT>
class ThrowableType : public ThrowableT
{
	public:
	ThrowableType() throw() 
	: ThrowableT(TagT::name(), 
			typename ThrowableT::DontManipulateStr()) {}

	explicit ThrowableType(const std::string& msg) throw()
	: ThrowableT(TagT::name() + ":\n  message:\n" + util::addIndent(msg, 4), 
			typename ThrowableT::DontManipulateStr()) {}

	ThrowableType(const char* filename, int linenumber) throw()
	: ThrowableT(ThrowableT::internalStrFromFilenameAndLinenumber(TagT::name(), filename, linenumber), 
			typename ThrowableT::DontManipulateStr()) {}

	ThrowableType(const char* filename, int linenumber, const char *function) throw()
	: ThrowableT(
		ThrowableT::internalStrFromFilenameLinenumberAndFunction(
			TagT::name(), filename, linenumber, function),
		typename ThrowableT::DontManipulateStr()) {}	

	ThrowableType(const char* filename, int linenumber, const std::string& msg) throw()
	: ThrowableT(
		ThrowableT::internalStrFromFilenameLinenumberAndMessage(TagT::name(), filename, linenumber, msg), 
			typename ThrowableT::DontManipulateStr()) {}

	ThrowableType(const char *filename, int linenumber, const char *function, const std::string& msg) throw()
	: ThrowableT(
		ThrowableT::internalStrFromFilenameLinenumberFunctionAndMessage(
			TagT::name(), filename, linenumber, function, msg),
		typename ThrowableT::DontManipulateStr()) {}	

	explicit ThrowableType(const std::exception& cause) throw()
	: ThrowableT(ThrowableT::internalStrFromCause(TagT::name(), cause), 
		typename ThrowableT::DontManipulateStr()) {}

	ThrowableType(const std::exception& cause, const std::string& msg) throw()
	: ThrowableT(ThrowableT::internalStrFromCauseAndMessage(TagT::name(), cause, msg),
		typename ThrowableT::DontManipulateStr()) {}	

	ThrowableType(const char *filename, int linenumber, const std::exception& cause) throw()
	: ThrowableT(ThrowableT::internalStrFromFilenameLinenumberAndCause(TagT::name(), filename, linenumber, cause), 
			typename ThrowableT::DontManipulateStr()) {}

	ThrowableType(const char *filename, int linenumber, const char *function, const std::exception& cause) throw()
	: ThrowableT(
		ThrowableT::internalStrFromFilenameLinenumberFunctionAndCause(
			TagT::name(), filename, linenumber, function, cause),
		typename ThrowableT::DontManipulateStr()) {}

	ThrowableType(const char *filename, int linenumber, const char *function,
		const std::exception& cause, const std::string& msg) throw()
	: ThrowableT(
		ThrowableT::internalStrFromFilenameLinenumberFunctionCauseAndMessage(
			TagT::name(), filename, linenumber, function, cause, msg),
		typename ThrowableT::DontManipulateStr()) {}

	virtual ~ThrowableType() throw() {}

	protected:
	ThrowableType(const std::string& msg, const typename ThrowableT::DontManipulateStr& dummy) throw()
	: ThrowableT(msg, dummy) {}	

};
}//namespace
#endif
