/* libthrowable - a C++ library to ease error handling and debugging
 *
 * Copyright (C) Matthias Langer 2006 <mlangc@gmx.at>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "trace_point.h"
#include "libthrowable/internal/internal.h"
#include "libthrowable/macros/here.h"
#include <iostream>
#include <map>
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

// LIBTHROWABLE_NO_THREADS may be defined to switch of
// multi-thread support even though it would be available.
#if !defined LIBTHROWABLE_NO_THREADS && (defined HAVE_PTHREAD || defined LIBTHROWABLE_WIN32)
#	define LIBTHROWABLE_THREADS
#	if defined HAVE_PTHREAD
#		include <pthread.h>
#	elif defined LIBTHROWABLE_WIN32
#		include <windows.h>
#	endif
#else
#	ifndef LIBTHROWABLE_NO_THREADS
#		define LIBTHROWABLE_NO_THREADS
#	endif
#endif

using namespace std;
using namespace libthrowable::util;
namespace libthrowable {
namespace {
namespace local {
internal::StringAlignmentOptimizer& getAlignmentOptimizer(internal::StringAlignmentOptimizer *p_optimizer = NULL) throw();

internal::StringAlignmentOptimizer& getAlignmentOptimizer(internal::StringAlignmentOptimizer *p_optimizer) throw()
{
	static internal::StringAlignmentOptimizer *sp_op = NULL;
	if(p_optimizer != NULL)
	{
		if(sp_op != NULL)
			delete sp_op;
		sp_op = p_optimizer;
	}
	else if(p_optimizer == NULL && sp_op == NULL)
		sp_op = new internal::StringAlignmentOptimizer(getTerminalWidth());
	return *sp_op;
}

#ifdef LIBTHROWABLE_THREADS
void checkReturn(int retVal, const char* funcName, const char* filename, int line, const char* function) throw()
{
	int errnum = retVal;
	#ifdef LIBTHROWABLE_WIN32
	errnum = GetLastError();
	#endif
	#if defined HAVE_PTHREAD
	if(retVal)
	#elif defined LIBTHROWABLE_WIN32
	if(!retVal)
	#else
	if(false)
	#endif
	{
		stringstream err;
		err << funcName << "(...) returned ";
		#ifdef LIBTHROWABLE_WIN32
		err << "FALSE";
		#else
		err << retVal;
		#endif
		err << endl;
		internal::emergencyAbort(err.str(), filename, line, function, errnum);
	}
}

template<class T> void dtorFunc(void* p_data)
{
	T* ptr = static_cast<T*>(p_data);
	delete ptr;
}

class NonCopyable
{
	protected:
	NonCopyable() {}
	~NonCopyable() {}
	
	private:
	NonCopyable(const NonCopyable&);
	NonCopyable& operator=(const NonCopyable&);
};

#if defined HAVE_PTHREAD
class PThreadMutex : private NonCopyable
{
	public:
	PThreadMutex() throw()
	{
		pthread_mutexattr_t mutexAttr;

		int rp = pthread_mutexattr_init(&mutexAttr);
		checkReturn(rp, "pthread_mutexattr_init", LIBTHROWABLE_HERE);

		rp = pthread_mutexattr_settype(&mutexAttr, PTHREAD_MUTEX_RECURSIVE);
		checkReturn(rp, "pthread_mutexattr_settype", LIBTHROWABLE_HERE);

		rp = pthread_mutex_init(&_mutex, &mutexAttr);
		checkReturn(rp, "pthread_mutex_init", LIBTHROWABLE_HERE);

		rp = pthread_mutexattr_destroy(&mutexAttr);
		checkReturn(rp, "pthread_mutexattr_destroy", LIBTHROWABLE_HERE);
	}

	~PThreadMutex() throw()
	{
		int rd = pthread_mutex_destroy(&_mutex);
		checkReturn(rd, "pthread_mutex_destroy", LIBTHROWABLE_HERE);
	}

	void lock() throw()
	{
		int rl = pthread_mutex_lock(&_mutex);
		checkReturn(rl, "pthread_mutex_lock", LIBTHROWABLE_HERE);
	}

	void unlock() throw()
	{
		int ru = pthread_mutex_unlock(&_mutex);
		checkReturn(ru, "pthread_mutex_unlock", LIBTHROWABLE_HERE);
	}

	private:
	pthread_mutex_t _mutex;
};

class PThreadID
{
	public:
	typedef pthread_t id_t;
		
	PThreadID(bool initialize=true) throw()
	: _id(0)
	{
		if(initialize)
			_id = pthread_self();
	}

	const id_t& get() const throw()
	{ return _id; }

	bool operator==(const PThreadID& id) const throw()
	{ return pthread_equal(_id, id._id); }

	bool operator!=(const PThreadID& id) const throw()
	{ return !operator==(id); }

	private:
	id_t _id;
};

template<class T> class ThreadSpecificPointer : private NonCopyable
{
	public:
	ThreadSpecificPointer() throw()
	{
		int rc = pthread_key_create(&_key, &dtorFunc<T>);
		checkReturn(rc, "pthread_key_create", LIBTHROWABLE_HERE);
	}

	~ThreadSpecificPointer() throw()
	{
		int rd = pthread_key_delete(_key);
		checkReturn(rd, "pthread_key_delete", LIBTHROWABLE_HERE);
	}

	void reset(T* p_data) throw()
	{
		if(get() || !p_data)
			dtorFunc<T>(get());
		int rs = pthread_setspecific(_key, p_data);
		checkReturn(rs, "pthread_setspecific", LIBTHROWABLE_HERE);
	}

	T* get() const throw()
	{ return static_cast<T*>(pthread_getspecific(_key)); }

	private:
	pthread_key_t _key;
};

typedef PThreadMutex Mutex;
typedef PThreadID ThreadID;

#elif defined LIBTHROWABLE_WIN32
// required for WinThreadSpecificPointer
class ScopedLock;

class WinMutex : private NonCopyable
{
	public:
	WinMutex() throw()
	{ InitializeCriticalSection(&_criticalSection); }
	~WinMutex() throw()
	{ DeleteCriticalSection(&_criticalSection); }

	void lock() throw()
	{ EnterCriticalSection(&_criticalSection); }

	void unlock() throw()
	{ LeaveCriticalSection(&_criticalSection); }

	private:
	CRITICAL_SECTION _criticalSection;
};

class WinThreadID
{
	public:
	typedef DWORD id_t;
		
	WinThreadID(bool initialize=true) throw()
	: _id(0)
	{
		if(initialize)
			_id = GetCurrentThreadId();
	}

	bool operator==(const WinThreadID& id) const throw()
	{ return _id == id._id; }
	bool operator!=(const WinThreadID& id) const throw()
	{ return !operator==(id); }

	const id_t& get() const throw()
	{ return _id; }

	private:
	DWORD _id;
};

// darn! this class cannot be called WinThreadSpecificPointer,
// because there are no template typedefs yet...
template<class T> class ThreadSpecificPointer : private NonCopyable
{
	public:
	ThreadSpecificPointer() throw()
	{ _key = TlsAlloc(); }

	~ThreadSpecificPointer() throw()
	{
		ScopedLock lock(_mutex);
		typename map<DWORD, T*>::const_iterator mi(_ptrs.begin());
		for(; mi != _ptrs.end(); ++mi)
			dtorFunc<T>(mi->second);
	}

	void reset(T* p_data) throw()
	{
		if(get() || !p_data)
			dtorFunc<T>(get());
		{
			ScopedLock lock(_mutex);
			_ptrs[GetCurrentThreadId()] = p_data;
		}
		TlsSetValue(_key, p_data);
	}

	T* get() const throw()
	{ return static_cast<T*>(TlsGetValue(_key)); }

	DWORD _key;
	map<DWORD, T*> _ptrs;
	WinMutex _mutex;
};

typedef WinMutex Mutex;
typedef WinThreadID ThreadID;
#endif 

Mutex tracePointMutex;
Mutex threadSwitchDetectorMutex;

class ScopedLock
{
	public:
	ScopedLock(Mutex& mutex) : _mutex(mutex)
	{ _mutex.lock(); }
	~ScopedLock()
	{ _mutex.unlock(); }

	private:
	Mutex& _mutex;
};

class ThreadSwitchDetector
{
	public:
	static bool changed(const ThreadID& tId) throw()
	{
		ScopedLock lock(threadSwitchDetectorMutex);
		static bool initialized = false;
		static ThreadID id(false);

		bool ret = !initialized;
		if(ret)
		{
			id = tId;
			initialized = true;
		}
		else
		{
			ret = (id != tId);
			if(ret)
				id = tId;
		}
		return ret;
	}

	private:
};
#endif //LIBTHROWABLE_THREADS
}}//local stuff

TracePoint::TracePoint(const char *filename, int linenumber, const char *function) throw()
: _function(filename, function)
{
	#ifdef LIBTHROWABLE_THREADS
	local::ScopedLock lock(local::tracePointMutex);
	#endif
	string output;
	genBasicOutput(output, linenumber);
	getOstream() << local::getAlignmentOptimizer().optimize(output) << endl;
}

TracePoint::TracePoint(const char *filename, int linenumber, const char *function, const string& message) throw()
: _function(filename, function)	
{
	#ifdef LIBTHROWABLE_THREADS
	local::ScopedLock lock(local::tracePointMutex);
	#endif
	string output;
	genBasicOutput(output, linenumber);
	if(!message.empty())
	{
		output += "\n";
		output += addIndent(message, 2*getLevel()+3);
	}
	getOstream() << local::getAlignmentOptimizer().optimize(output) << endl;
}

TracePoint::~TracePoint() throw()
{
	fStackPop();
}

void TracePoint::genBasicOutput(string& toWriteIn, int linenumber) const throw()
{
	#ifdef LIBTHROWABLE_THREADS
	local::ThreadID id;
	if(local::ThreadSwitchDetector::changed(id))
	{
		string tmp1("# thread " + stringify(id.get()));
		toWriteIn += addIndent(tmp1, 2*getLevel()) + "\n";
	}
	#endif
	
	if(fStackPush())
	{
		string tmp2("+ " + _function.name);
		toWriteIn += addIndent(tmp2, 2*(getLevel()-1)) + "\n";
	}
	string tmp3 = "* trace point: " + _function.file + ":" + stringify(linenumber);
	toWriteIn += addIndent(tmp3, 2*getLevel());
}

void TracePoint::genVarOutput(string& toWriteIn, int linenumber, const string& varStr, const string& varName) const throw()
{
	genBasicOutput(toWriteIn, linenumber);
	toWriteIn += "\n";
	toWriteIn += addIndent(varName.empty() ? "<unknown variable>" : varName, 2*getLevel()+3);
	// if an object's string representation contains newlines,
	// add another 2 spaces of indent and use a slightly different
	// format 
	// (0x0A = newline)
	if(varStr.find(0x0A) != string::npos)
	{
		toWriteIn += ":\n";
		toWriteIn += addIndent(varStr, 2*getLevel()+5);
	}
	else
		toWriteIn += (" = " + varStr);
}

bool TracePoint::fStackPush() const throw()
{
	bool ret = false;
	stack<Function>& fStack = getFunctionStack();
	if(fStack.empty() || fStack.top() != _function)
	{
		++getLevel();
		ret = true;
	}
	fStack.push(_function);
	return ret;
}

void TracePoint::fStackPop() const throw()
{
	stack<Function>& fStack = getFunctionStack();
	fStack.pop();
	if(fStack.empty() || fStack.top() != _function)
		--getLevel();
}

//static members
void TracePoint::setLineWidth(int width) throw()
{
	if(width < 0)
		local::getAlignmentOptimizer(new internal::StringAlignmentOptimizer(getTerminalWidth()));
	else
		local::getAlignmentOptimizer(new internal::StringAlignmentOptimizer(width));
}

void TracePoint::setOstream(ostream& os) throw()
{ getOstream(&os); }

ostream& TracePoint::getOstream(ostream *p_os) throw()
{
	static ostream *sp_os = &cerr;
	if(p_os != NULL)
		sp_os = p_os;
	return *sp_os;
}

stack<TracePoint::Function>& TracePoint::getFunctionStack() throw()
{
	#ifndef LIBTHROWABLE_THREADS
	static stack<Function> s_stack;
	return s_stack;
	#else
	static local::ThreadSpecificPointer<stack<Function> > threadSpecificPtr;
	if(!threadSpecificPtr.get())
		threadSpecificPtr.reset(new stack<Function>());
	return *threadSpecificPtr.get();
	#endif
}

unsigned& TracePoint::getLevel() throw()
{
	#ifndef LIBTHROWABLE_THREADS
	static unsigned s_level;
	return s_level;
	#else
	static local::ThreadSpecificPointer<unsigned> threadSpecificPtr;
	if(!threadSpecificPtr.get())
		threadSpecificPtr.reset(new unsigned(0));
	return *threadSpecificPtr.get();
	#endif	
}

string TracePoint::optimizeForLineWidth(const string& str) throw()
{ return local::getAlignmentOptimizer().optimize(str); }
}//namespace	

