/* libthrowable - a C++ library to ease error handling and debugging
 *
 * Copyright (C) Matthias Langer 2006 <mlangc@gmx.at>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef H_LIBTHROWABLE_TRACER
#define H_LIBTHROWABLE_TRACER
#include "util/util.h"
#include <ostream>
#include <stack>
namespace libthrowable {
	/** A class that can be very useful for finding bugs.
	 * @ingroup debugging macros
	 * @warning
	 * 	If you are using TracePoint in a multi threaded environment, based on
	 * 	pthreads, make sure that the library has been compiled with
	 * 	pthreads support. This can be accomplished by passing
	 * 	"--enable-pthreads" to configure. If you are using the Gentoo ebuild,
	 * 	be sure that you emerge libthrowable with USE="threads".
	 * @see #LIBTHROWABLE_TRACE
	 * @see #LIBTHROWABLE_TRACE_WITH_MSG
	 * @since libthrowable-0.9.1*/
	class LIBTHROWABLE_API TracePoint
	{
		public:
		/** Constructs a new TracePoint object from filename, linenumber and function.
		 * @see TracePoint(const char*, int, const char*, const std::string&)
		 * @see #LIBTHROWABLE_TRACE*/
		TracePoint(const char *filename, int linenumber, const char *function) throw();

		/** Constructs a new TracePoint object from filename, linenumber, function and message.
		 * @see TracePoint(const char*, int, const char*)
		 * @see #LIBTHROWABLE_TRACE_WITH_MSG*/
		TracePoint(const char *filename, int linenumber, const char *function, const std::string& message) throw();

		/** Constructs a new TracePoint object from filename, linenumber, function, a variable and
		 *  its name.
		 * @param filename
		 * 	usually __FILE__
		 * @param linenumber
		 * 	usually __LINE__
		 * @param function
		 * 	usually __FUNCTION__, __func__ or __PRETTY_FUNCTION__
		 * @param t 
		 * 	any type meeting the requirements for util::stringify()
		 * @param name 
		 * 	the name of the variable you want to examine
		 * @see #LIBTHROWABLE_TRACE_VAR
		 * @since libthrowable-1.0.0
		 */
		template<class OStreamableT> TracePoint(const char *filename, int linenumber, const char *function, 
		                                         const OStreamableT& t, const std::string& name) throw()
		: _function(filename, function)
		{ 
			std::string output;
			genVarOutput(output, linenumber, util::stringify(t), name);
			getOstream() << optimizeForLineWidth(output) << std::endl;
		}

		~TracePoint() throw();
			
		/** Specify a line width for all TracePoint objects.
		 * @param width
		 * <ul>
		 * 	<li>
		 * 		If width < 0 the output will be optimized for your current terminal.
		 *		This is the default.
		 *	</li>
		 *	<li>
		 * 		If 0 <= width < 40 no optimizing will be done.
		 * 	</li>
		 * 	<li>
		 * 		If width <= 40 all TracePoint objects will generate output optimized for width.
		 * 		This is mostly useful when writing to files.
		 * 	</li>
		 * </ul>*/	
		static void setLineWidth(int width) throw();

		/** Write to os instead of std::cerr.
		 * @warning
		 * 	The caller of this function is responsible for making sure
		 * 	that os remains a valid reference as long as used by various
		 * 	trace points.*/
		static void setOstream(std::ostream& os) throw();

		private:
		//no copy, no assignement
		TracePoint(const TracePoint& other);
		TracePoint& operator=(const TracePoint& other);

		struct Function
		{
			Function(const char *tFile, const char *tName) throw()
			: file(tFile), name(tName) {}

			bool operator==(const Function& other) const throw()
			{ return (name == other.name && file == other.file); }

			bool operator!=(const Function& other) const throw()
			{ return !(*this == other); }

			const std::string file;
			const std::string name;
		};

		//static
		static std::ostream& getOstream(std::ostream *p_os = NULL) throw();
		static std::stack<Function>& getFunctionStack() throw();

		static unsigned& getLevel() throw();
		static std::string optimizeForLineWidth(const std::string& str) throw();
		////////

		void genBasicOutput(std::string& toWriteIn, int linenumber) const throw();
		void genVarOutput(std::string& toWriteIn, int linenumber, const std::string& varStr, const std::string& varName) const throw();

		//pushes _function on the stack and
		//	increases _s_level if we have just entered _function and returns true.
		//	returns false otherwise
		bool fStackPush() const throw();

		//pops from the stack and
		//	decreases _s_level if we just left _function
		void fStackPop() const throw();

		const Function _function;

	};
}//namespace
#endif
