/* libthrowable - a C++ library to ease error handling and debugging
 *
 * Copyright (C) Matthias Langer 2006 <mlangc@gmx.at>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef H_LIBTHROWABLE_USER_FAULT
#define H_LIBTHROWABLE_USER_FAULT
#include "exception.h"
#include "system.h"
namespace libthrowable {
/** This class describes an Exception object directly aimed at the enduser.
 * @ingroup exceptions
 * @note
 * 	'an Exception object aimed directly at the enduser' should be considered
 *  as error message to be consumed by the user boxed in an Exception object.
 *  Imagine for example you write a calculator. When your (correctly working) parse method fails
 *  while parsing it's input, it is most likely, that the user, who indirectly
 *  supplied it with an invalid expression is responsible for this problem,
 *  so he/she is the one who should 'catch' your exception. Thus you will do something
 *  like
 *  @code
 *  	if(parsing_failed)
 *  		throw UserFault("not a valid expression: " + expr);
 *  @endcode
 *  The client code of your calculator class may look like this:
 *  @code
 *  	try
 *  	{
 *  		double result = calculator.parseExpr(expr);
 *			cout << "expr = " << result;
 *		}
 *		catch(const UserFault& fault)
 *		{
 *			cout << fault << endl;
 *		}
 *  @endcode 
 */	
class LIBTHROWABLE_API UserFault : public Exception
{
	public:
	/** Constructs a new UserFault from msg.
	 * @param msg
	 * 	This should be an (possibly) translated error message for the user,
	 * 	not for you as programmer.*/
	explicit UserFault(const std::string& msg) throw();

	/** Constructs a new UserFault from filename, linenumber and msg.*/
	UserFault(const char *filename, int linenumber, const std::string& msg) throw();

	/** Constructs a new UserFault from filename, linenumber, function and msg.*/
	UserFault(const char *filename, int linenumber, const char *function, const std::string& msg) throw();

	/** Constructs a new UserFault from cause.*/
	UserFault(const std::exception& cause, const std::string& msg) throw();

	/** Returns an error message targeted at the end user.*/
	virtual const std::string& getMessage() const throw();

	virtual ~UserFault() throw();

	protected:
	/** You may want to use this constructor when derivin classes from this one.
	 * @param msg
	 * 	This string defines the return value of getMessage() as long as you don't overide this method.
	 * @param passUnmodifiedToException
	 * 	This string defines the return value of what() as long as you don't override this method.*/
	UserFault(const std::string& msg, const std::string& passUnmodifiedToException) throw();

	/** Gives you a refernce to the string that defines the return value of getMessage().
	 * @note
	 * 	You should rarely use this method.*/
	std::string& getInternalMessage() throw();

	private:
	std::string _message;
};
std::ostream& operator<<(std::ostream& os, const UserFault& uf);
}//namespace
#endif
