/* libthrowable - a C++ library to ease error handling and debugging
 *
 * Copyright (C) Matthias Langer 2006 <mlangc@gmx.at>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <libthrowable.h>
#include <libthrowable/internal/internal.h>
#include <iostream>
#include <sstream>
#ifdef HAVE_PTHREAD
#include <pthread.h>
#endif
#ifdef LIBTHROWABLE_WIN32
#include <windows.h>
#include <conio.h>
#endif

#if defined __GNUC__
#define LIBTHROWABLE_FUNCTION __PRETTY_FUNCTION__
#elif defined _MSC_VER
#define LIBTHROWABLE_FUNCTION __FUNCSIG__
#else
#define LIBTHROWABLE_FUNCTION __FUNCTION__
#endif


using namespace libthrowable;
using namespace libthrowable::util;
using libthrowable::internal::StringAlignmentOptimizer;
using namespace std;

namespace {
namespace local {

class MultiLineOutput
{
	public:
	MultiLineOutput(int a, int b)
	: _a(a), _b(b) {}
	friend ostream& operator<<(ostream& os, const MultiLineOutput& multiLineOutput)
	{
		os << " ________ " << endl
		   << "/ __  __ \\" << endl
		   << "|   |    |" << endl
		   << "| \\____/ |" << endl
		   << "\\________/" << endl;
		os << endl;

		const int& a = multiLineOutput._a;
		const int& b = multiLineOutput._b;
		os << a << " * " << b << endl;
		os << "= " << a*b << endl;
		return os;

	}

	private:
	int _a, _b;
};

struct Match
{
	Match() 
	: full(true), sizeDiff(false), firstDiff(string::npos) {}
	
	Match(bool tSizeDiff, unsigned tFirstDiff) 
	: full(false), sizeDiff(tSizeDiff), firstDiff(tFirstDiff) {}
	
	const bool full;
	const bool sizeDiff;
	const unsigned firstDiff;
};

Match strMatch(const string& str1, const string& str2)
{
	bool sizeDiff = false;
	if(str1.size() != str2.size())
		sizeDiff = true;
	unsigned mSize = (str1.size() < str2.size()) ? str1.size() : str2.size();
	for(unsigned i=0; i != mSize; ++i)
	{
		if(str1[i] != str2[i])
			return Match(sizeDiff, i);
	}
	if(sizeDiff)
		return Match(sizeDiff, string::npos);
	return Match();
}

void fail(int linenumber, const Throwable& th, const string& expected)
{
	cerr << "  failure at " <<  __FILE__ << ":" << linenumber << endl;
	cerr << "    caught:" << endl;
	cerr << th.what(6) << endl;
	cerr << "    expected:" << endl;
	cerr << addIndent(expected, 6) << endl << endl;
}

void  LIBTHROWABLE_NORETURN failInMain();
void failInMain()
{
	cerr << "[!!]" << endl;
	exit(1);
}

void succeed(int linenumber, const Throwable& th)
{
	cout << "  success at " << __FILE__ << ":" << linenumber << endl;
	cout << "    caught:" << endl;
	cout << th.what(6) << endl << endl;
}

bool testBaseFunctionality()
{
	int linenumber;
	string message;
	try
	{
		message = 
			"problems at block #4:\n"
			"  consider emergency cutout";

		linenumber = __LINE__ + 1;
		throw Problem(__FILE__, __LINE__, message);
	}
	catch(const Throwable& th)
	{
		string expected = 
			"Problem:\n"
			"  where: " + string(__FILE__) + ":" + stringify(linenumber) + "\n"
			"  message:\n" + addIndent(message, 4);
		if(expected != th.what())
		{
			fail(__LINE__, th, expected);
			return false;
		}
		else
			succeed(__LINE__, th);
	}

	Problem problem;
	try
	{
		string nestedMessage;
		try
		{
			nestedMessage = 
				"block 4:\n"
				"  you are violating safety regulations:\n"
				"    power output 30 MW is too low";
			throw Problem(nestedMessage);
		}
		catch(const Problem& pr)
		{
			string expected =
				"Problem:\n"
				"  message:\n" +
				addIndent(nestedMessage, 4);
			if(expected != pr.what())
			{
				fail(__LINE__, pr, expected);
				return false;
			}
			else
			{
				succeed(__LINE__, pr);
				problem = pr; 
				linenumber = __LINE__ + 1;
				throw Problem(__FILE__, __LINE__, pr);
			}
		}
	}
	catch(const Throwable& th)
	{
		string expected =
			"Problem:\n"
			"  where: " +
			string(__FILE__) + ":" + stringify(linenumber) + "\n" +
			"  cause:\n" +
			problem.what(4);
		if(expected != th.what())
		{
			fail(__LINE__, th, expected);
			return false;
		}
		else
			succeed(__LINE__, th);
	}
	return true;
}

bool testMacros()
{
	int linenumber;
	try
	{
		linenumber = __LINE__ + 1;
		throw Exception(LIBTHROWABLE_HERE);
	}
	catch(const Exception& e)
	{
		string expected =
			"Exception:\n"
			"  where: " + string(__FILE__) + ":" + stringify(linenumber) + " in\n    " + string(LIBTHROWABLE_FUNCTION);
		if(expected != e.what())
		{
			fail(__LINE__, e, expected);
			return false;
		}
		else
			succeed(__LINE__, e);
	}
	
	string message =
		"FATAL:\n"
		"  reactor output increasing too fast\n"
		"  output right now:\n"
		"    30 GW";
	
	try
	{
		linenumber = __LINE__ + 1;
		 throw Error(LIBTHROWABLE_HERE,  message);
	}
	catch(const Error& err)
	{
		string expected =
			"Error:\n"
			"  where: " + string(__FILE__) + ":" + stringify(linenumber) + " in\n    " + string(LIBTHROWABLE_FUNCTION) + "\n" +
			"  message:\n" +
			addIndent(message, 4);

		if(expected != err.what())
		{
			fail(__LINE__, err, expected);
			return false;
		}
		else
			succeed(__LINE__, err);
	}

	message = 
		"assertion \"rodInserted == true\" failed !";
	try
	{
		bool rodInserted = false;
		linenumber = __LINE__ + 1;
		LIBTHROWABLE_ASSERT(rodInserted == true);
	}
	catch(const FailedAssertion& failedAssertion)
	{
		string expected =
			"FailedAssertion:\n"
			"  where: " + string(__FILE__) + ":" + stringify(linenumber) + " in\n    " + string(LIBTHROWABLE_FUNCTION) + "\n" +
			"  message:\n" +
			addIndent(message, 4);
		if(expected != failedAssertion.what())
		{
			fail(__LINE__, failedAssertion, expected);
			return false;
		}
		else
			succeed(__LINE__, failedAssertion);
	}
	
	try
	{
		#define LIBTHROWABLE_NDEBUG
		LIBTHROWABLE_ASSERT("1 == 2");
		#undef LIBTHROWABLE_NDEBUG
	}
	catch(const FailedAssertion& failedAssertion)
	{
		cerr << "  failure at " + string(__FILE__) + stringify(__LINE__);
		cerr << "    LIBTHROWABLE_NDEBUG did not work !" << endl;
		return false;
	}

	try
	{
		try
		{
			linenumber = __LINE__ + 1;
			 throw Problem(LIBTHROWABLE_HERE,  string("Radiation level critical !"));
		}
		catch(const Throwable& th)
		{
			throw Problem(th, "EMERGENCY: EVACUATE IMMEDEATLY !");
		}
	}
	catch(const Throwable& th)
	{
		string expected =
			"Problem:\n"
			"  message:\n"
			"    EMERGENCY: EVACUATE IMMEDEATLY !\n"
			"  cause:\n"
			"    Problem:\n"
			"      where: " + string(__FILE__) + ":" + stringify(linenumber) + " in\n"
			"        " + string(LIBTHROWABLE_FUNCTION) + "\n"
			"      message:\n"
			"        Radiation level critical !";
		
		if(expected != th.what())
		{
			fail(__LINE__, th, expected);
			return false;
		}
		else
			succeed(__LINE__, th);
	}
	
	try
	{
		try
		{
			throw Error(string("Cooling system seriously damaged !"));
		}
		catch(const Throwable& th)
		{
			linenumber = __LINE__ + 1;
			throw Error(LIBTHROWABLE_HERE, th, "Emergency shutdown failed !");
		}
	}
	catch(const Throwable& th)
	{
		string expected =
			"Error:\n"
			"  where: " + string(__FILE__) + ":" + stringify(linenumber) + " in\n"
			"    " + string(LIBTHROWABLE_FUNCTION) + "\n"
			"  message:\n"
			"    Emergency shutdown failed !\n"
			"  cause:\n"
			"    Error:\n"
			"      message:\n"
			"        Cooling system seriously damaged !";

		if(expected != th.what())
		{
			fail(__LINE__, th, expected);
			return false;
		}
		else
			succeed(__LINE__, th);
	}

	return true;
}

bool testToString()
{
	string msg =
		"This is a story about spaces, tabs and newlines and small terminals:\n"
		"  There is a problem when writing output to a terminal:\n"
		"    The problem is the width of the terminal:\n"
		"      Will this line really fit into one line in your current window ? "
			  	"What if not ?\n"
		"        Without further measures, reading this story may be a pain in the ass, "
		  			"just because your terminal is too small.\n"
		"        And what should we do, if someone really wants to start his output after say "
					"68 spaces ? Confused ? Let's take a look:\n"
		+ string(68, ' ') +
		"Yes, what you read here has an indent of 68 (+ 4) spaces. Not enough ?\n"
		+ string(72, ' ') +
		"Well, here are 72 (+ 4) ... \n"
		+ string(83, ' ') +
		"or even 83 (+ 4) ...\n"
		+ string(666, ' ') +
		"666 (+ 4)\n"
		+ string(672, ' ') +
		"672 (+ 4)\n"
		+ string(666, ' ') +
		"666 (+ 4) again\n" 
		+ string(664, ' ') +
		"now 664 (+ 4)\n"
		+ string(42, ' ') +
		"why not try 42 (+ 4) ... somebody once said, that 42 is the answer to all questions.\n"
		+ string(43, ' ') +
		"to be onest, nobody ever told be, why 43 is a bad answer\n"
		+ string(41, ' ') +
		"or 41 is wrong\n"
		"       However, i think you know now what mean; of course, the same applies to tabs, after "
			"they have been converted to spaces:"
		+ string(3, '\t') +
		"what i tell you here, has an indent of 3 tabs\n"
		+ string(4, '\t') +
		"or should i better use 4 ?\n"
		+ string(23, '\t') +
		"or even 23 ? You know, there's a movie out there called 23. Some people say 23 is " 
		"the number of conspiracy. Sorry, if this scares you, but sometimes, i smoke "
		"cigarettes that are called 'Ernte 23'.\n"
		"    Are you still reading this crap ? If so, you're almost done because this is "
			"where this weird story ends.";
		 
	try
	{
		 throw Problem(LIBTHROWABLE_HERE,  msg);
	}
	catch(const Problem& pr)
	{
		cout << pr.toString() << endl;
	}
	return true;
}

//declare some user faults
namespace tags {
	struct InsaneInputTag
	{
		static const string& name()
		{
			static const string ret("InsaneInput");
			return ret;
		}
	};

	struct TooManyMouseClicksTag
	{
		static const string& name()
		{
			static const string ret("TooManyMouseClicks");
			return ret;
		}
	};

	struct OperationImpossibleTag
	{
		static const string& name()
		{
			static const string ret("OperationImpossible");
			return ret;
		}
	};
}//namespace tags
typedef UserFaultType<UserFault, tags::InsaneInputTag> InsaneInput;
typedef UserFaultType<InsaneInput, tags::TooManyMouseClicksTag> TooManyMouseClicks;
typedef UserFaultType<UserFault, tags::OperationImpossibleTag> OperationImpossible;

bool testUserFaultFromType()
{
	string message;
	int linenumber;
	try
	{
		message =
			"Only 10 clicks within one second are allowed !";
		linenumber = __LINE__ + 1;
		 throw TooManyMouseClicks(LIBTHROWABLE_HERE,  message);
	}
	catch(const UserFault& uf)
	{
		if(message != uf.getMessage())
		{
			cerr << "  failure at " << __FILE__ << ":" << __LINE__ << endl;
			cerr << "    message:" << endl;
			cerr << addIndent(message, 6) << endl;
			cerr << "    uf.getMessage():" << endl;
			cerr << addIndent(uf.getMessage(), 6) << endl << endl;
			return false;
		}
		else
		{
			cout << "  success at " << __FILE__ << ":" << __LINE__ << endl;
			cout << "    uf.getMessage():" << endl;
			cout << addIndent(uf.getMessage(), 6) << endl << endl;
		}
	}
	for(unsigned i=0; i != 3; ++i)
	{
		try
		{
			if(i == 0)
			{
				message = 
					"Please don't click the mouse that often !\n"
					"This may confuse the program.";
				linenumber = __LINE__ + 1;
				 throw TooManyMouseClicks(LIBTHROWABLE_HERE,  message);
			}
			else if(i == 1)
			{
				message =
					"\":cheat\": argument expected\n"
					"try \":help cheat\"";
				linenumber = __LINE__ + 1;
				 throw InsaneInput(LIBTHROWABLE_HERE,  message);
			}
			else
			{
				message =
					"Sorry, \"66.666.66.66\" could not be found !\n"
					"If you are sure this is a valid ip number, try again later.";
				linenumber = __LINE__ + 1;
				 throw UserFault(LIBTHROWABLE_HERE,  message);
			}
		}
		catch(const TooManyMouseClicks& tmmc)
		{
			if(i != 0)
			{
				cerr << "  failure at " << __FILE__ << ":" << __LINE__ << endl << endl;
				return false;
			}
			else
				succeed(__LINE__, tmmc);
		}
		catch(const InsaneInput& ii)
		{
			if(i != 1)
			{
				cerr << "  failure at " << __FILE__ << ":" << __LINE__ << endl << endl;
				return false;
			}
			else
				succeed(__LINE__, ii);
		}
		catch(UserFault& uf)
		{
			if(i !=2)
			{
				cerr << "  failure at " << __FILE__ << ":" << __LINE__ << endl << endl;
				return false;
			}
			else
				succeed(__LINE__, uf);
		}
	}
	return true;
}

bool testStringAlignmentOptimizer()
{
	string tStr =
		"This is something we want to look good, even with a linewidth of 60. "
		"You may notice that this is not possible without further invention, "
		"because till now, we didn't do a line break manually. "
		"Instead, we are hoping that the neccessairy linebreaks will be inserted "
		"for us.\n" 
		+ string(40, ' ') +
		"For testing purposes we contiue with an indent of 40:\n"
		+ string(40, ' ') + 
		"  As you may notice 40 = 8*5 = 2*2*2*5.\n"
		+ string(60, ' ') +
		"What about 60 ?\n" 
		+ string(60, ' ') +
		"  Well, 60 = 6*10 = 2*3*2*5 = 2*2*3*5.\n"
		+ string(220, ' ') +
		"And 220 ?\n" 
		+ string(220, ' ') +
		"  Why not:\n"
		+ string(220, ' ') +
		"    220 = 22*10 = 11*2*2*5 = 2*2*5*11.\n"
		+ string(68, ' ') +
		"Back to 68:\n" 
		+ string(68, ' ') +
		"  68 = 4*17 = 2*2*17.\n"
		+ string(5, ' ') +
		"What to do with 5 ?\n"
		+ string(5, ' ') +
		"  Are you kidding ? 5 = 5, of course :-)";

	string expected =
		"This is something we want to look good, even with a\n"
		"linewidth of 60. You may notice that this is not possible\n"
		"without further invention, because till now, we didn't do a\n"
		"line break manually. Instead, we are hoping that the\n"
		"neccessairy linebreaks will be inserted for us.\n"
		"--40--spaces-->\n"
		"For testing purposes we contiue with an indent of 40:\n"
		"  As you may notice 40 = 8*5 = 2*2*2*5.\n"
		"                    What about 60 ?\n"
		"                      Well, 60 = 6*10 = 2*3*2*5 = 2*2*3*5.\n"
		"--180--spaces-->\n"
		"And 220 ?\n"
		"  Why not:\n"
		"    220 = 22*10 = 11*2*2*5 = 2*2*5*11.\n"
		"<--180-spaces--\n"
		"                            Back to 68:\n"
		"                              68 = 4*17 = 2*2*17.\n"
		"<--40-spaces--\n"
		"     What to do with 5 ?\n"
		"       Are you kidding ? 5 = 5, of course :-)";

	StringAlignmentOptimizer optimizer(60);
	string result(optimizer.optimize(tStr));
	Match match(strMatch(expected, result));
	if(!match.full)
	{
		cerr << "failure at " << __FILE__ << ":" << __LINE__ << endl;
		cerr << "  match.sizeDiff: " << match.sizeDiff << endl;
		if(match.firstDiff != string::npos)
		{
			cerr << "  match.firstDiff: " << match.firstDiff << endl;
			cerr << "  result:" << endl;
			cerr << addIndent(result, 4) << endl << endl;
			cerr << "  expected:" << endl;
			cerr << addIndent(expected, 4) << endl << endl;
			cerr << "  matched till here:" << endl;
			cerr << addIndent(expected.substr(0, match.firstDiff), 4) << endl;
			return false;
		}
	}
	cout << result << endl;
	return true;
}



void f1();
void f2();
void f3();

void f3()
{
	LIBTHROWABLE_TRACE(tp1);
	for(unsigned i=0; i != 3; ++i)
		LIBTHROWABLE_TRACE_VAR(tp2, i);
	f1();
}

void f2()
{
	LIBTHROWABLE_TRACE(tp1);
	f3();
	LIBTHROWABLE_TRACE(tp2);
}

void f1()
{
	static unsigned recDepth = 0;
	if(recDepth == 2)
	{
		LIBTHROWABLE_TRACE_VAR(tp1, recDepth);
		return;
	}
	++recDepth;
	LIBTHROWABLE_TRACE(tp2);
	for(unsigned i=0; i != 2; ++i)
		f2();
	LIBTHROWABLE_TRACE(tp3);
}

bool testTracer()
{
	//verify that Tracer::setOstream(...)
	//at least doesn't segfault :-)
	TracePoint::setOstream(cout);
	LIBTHROWABLE_TRACE(tp1);
	TracePoint::setOstream(cerr);

	LIBTHROWABLE_TRACE_WITH_MSG(tp2, "once our TracePoint didn't like newlines:\n\n\n");
	f1();
	LIBTHROWABLE_TRACE(tp3);

	LIBTHROWABLE_TRACE_WITH_MSG(tp4, "testing multi line object output in TracePoint");
	MultiLineOutput multiLineOutput(719, 941);
	LIBTHROWABLE_TRACE_VAR(tp5, multiLineOutput);
	return true;
}

#if defined LIBTHROWABLE_WIN32 && !defined LIBTHROWABLE_CYGWIN 
// the test application would simply quit after
// finishing, allowing no examination of its
// output
void waitOnExit()
{
	cout << endl << "Press any key to quit" << flush;
	_getch();
}
#endif

/////////////////////
//test thread support
/////////////////////
#if defined HAVE_PTHREAD && !defined LIBTHROWABLE_NO_THREADS
class PThread
{
	template<class T> class CallbackWrapper
	{
		public:
		static void* call(void*)
		{
			T::call();
			return NULL;
		}
	};
	public:
	template<class T_FUNCTOR> void init()
	{
		int rc = pthread_create(&_thread, NULL, &CallbackWrapper<T_FUNCTOR>::call, NULL);
		LIBTHROWABLE_ASSERT(rc == 0);
	}

	void join()
	{
		int rj = pthread_join(_thread, NULL);
		LIBTHROWABLE_ASSERT(rj == 0);
	}

	static void exit()
	{
		pthread_exit(NULL);
	}

	private:
	pthread_t _thread;
};
typedef PThread Thread;

#elif defined LIBTHROWABLE_WIN32 && !defined LIBTHROWABLE_NO_THREADS
void sleep(unsigned seconds)
{
	// Sleep() uses milliseconds
	Sleep(seconds*1000);
}

class WinThread
{
	template<class T> class CallbackWrapper
	{
		public:
		static DWORD WINAPI call(void*)
		{
			T::call();
			return 0;
		}
	};
	public:
	WinThread() : _handle(NULL) {}

	~WinThread()
	{
		if(_handle)
			CloseHandle(_handle);
	}
	
	template<class T_FUNCTOR> void init()
	{
		_handle = CreateThread(NULL, 0, &CallbackWrapper<T_FUNCTOR>::call, NULL, 0, NULL);
		LIBTHROWABLE_ASSERT(_handle != NULL);
	}

	void join()
	{
		DWORD rw = WaitForSingleObject(_handle, INFINITE);
		LIBTHROWABLE_ASSERT(rw != WAIT_FAILED);
	}

	static void exit()
	{
		ExitThread(0);
	}

	private:
	HANDLE _handle;
};
typedef WinThread Thread;

#else
#	ifndef LIBTHROWABLE_NO_THREADS
#		define LIBTHROWABLE_NO_THREADS
#	endif
#endif

#ifndef LIBTHROWABLE_NO_THREADS
void threadSaveRecStuffStart(unsigned i);
void threadSaveRecStuff(unsigned i);

void threadSaveRecStuffStart(unsigned i)
{
	LIBTHROWABLE_TRACE(tp1);
	const unsigned breakOut = 2;
	if(i == breakOut)
	{
		LIBTHROWABLE_TRACE_WITH_MSG(tp2, "leaving recursion");
		return;
	}
	threadSaveRecStuff(i);
}

void threadSaveRecStuff(unsigned i)
{
	LIBTHROWABLE_TRACE(tp1);
	++i;
	threadSaveRecStuffStart(i);
}

/**
 * As the callback function interfaces for Windows and
 * pthreads differ (Windows: DWORD WINAPI function(void*),
 * pthreads: void* function(void*)), this pseudo functor
 * allows you to pass a callback function to the Thread::init()
 * function.
 * TODO: The approach below is not a really good one as you
 *       could as well pass a function pointer. Switch to
 *       using real objects, not pseudo functors.
 */
class ThreadFtor1
{
	public:
	static void call()
	{
		LIBTHROWABLE_TRACE(tp1);
		sleep(1);
		threadSaveRecStuffStart(0);
		LIBTHROWABLE_TRACE(tp2);
		Thread::exit();
	}
};

class ThreadFtor2
{
	public:
	static void call()
	{
		LIBTHROWABLE_TRACE(tp1);
		threadSaveRecStuffStart(0);
		sleep(1);
		LIBTHROWABLE_TRACE(tp2);
		Thread::exit();
	}
};


bool testTracerWithThreads()
{
	try
	{
		LIBTHROWABLE_TRACE_WITH_MSG(tp1, "just entered testTracerWithThreads");
		const unsigned threadMax = 4;
		Thread threads[threadMax];
		for(unsigned i=0; i != threadMax; ++i)
		{
			if(i % 2)
				threads[i].init<ThreadFtor1>();
			else
				threads[i].init<ThreadFtor2>();
		}
		for(unsigned i=0; i != threadMax; ++i)
			threads[i].join();
		
		LIBTHROWABLE_TRACE_WITH_MSG(tp2, "leaving testTracerWithThreads");
		return true;
	}
	catch(const Throwable& th)
	{
		cerr << "Throwable caught in testTracerWithThreads:" << endl;
		cerr << th << endl;
		return false;
	}
}
#endif
}}//local namespace

int main()
{
	#if defined LIBTHROWABLE_WIN32 && !defined LIBTHROWABLE_CYGWIN
	atexit(&local::waitOnExit);
	#endif

	cout << "testing base functionality:" << endl;
	if(local::testBaseFunctionality())
		cout << "[Ok]" << endl << endl;
	else
		local::failInMain();
	cout << "testing macros" << endl;
	if(local::testMacros())
		cout << "[Ok]" << endl << endl;
	else
		local::failInMain();
	if(local::testUserFaultFromType())
		cout << "[Ok]" << endl << endl;
	else
		local::failInMain();
	cout << "testing toString()" << endl;
	if(local::testToString())
		cout << "[Ok]" << endl << endl;
	else
		local::failInMain();
	cout << "testing StringAlignmentOptimizer ... " << endl;
	if(local::testStringAlignmentOptimizer())
		cout << "[Ok]" << endl << endl;
	else
		local::failInMain();
	cout << "testing TracePoint" << endl;
	if(local::testTracer())
		cout << "[Ok]" << endl;
	else
		local::failInMain();

	#ifndef LIBTHROWABLE_NO_THREADS
	cout << endl;
	cout << "testing TracePoint with threads ... " << endl;
	if(local::testTracerWithThreads())
		cout << "[Ok]" << endl;
	else
	{
		cout << "[!!]" << endl;
		return 77;
	}
	#endif
	return 0;
}
